export = getTokenBeforeClosingBracket;
/**
 * Find the token before the closing bracket.
 * @param {ASTNode} node - The JSX element node.
 * @returns {Token} The token before the closing bracket.
 */
declare function getTokenBeforeClosingBracket(node: ASTNode): Token;
//# sourceMappingURL=getTokenBeforeClosingBracket.d.ts.map