/**
 * Checks if we are declaring a `props` argument with a flow type annotation.
 * @param {ASTNode} node The AST node being checked.
 * @param {Object} context
 * @returns {boolean} True if the node is a type annotated props declaration, false if not.
 */
export function isAnnotatedFunctionPropsDeclaration(node: ASTNode, context: any): boolean;
//# sourceMappingURL=annotations.d.ts.map